/*
** mpeg.c
** 
** Made by (Arne Caspari)
** Login   <arne@arne-laptop>
** 
** Started on  Tue Feb 27 07:19:39 2007 Arne Caspari
** Last update Mon Mar 19 07:45:32 2007 Arne Caspari
*/

#include <avcodec.h>
#include <avformat.h>
#include <string.h>

#include "mpeg.h"
#include "unicap.h"
#define DEBUG
#include "debug.h"
#include "../libunicap/unicap_private.h"

static int initialized = 0;

#define INBUF_SIZE 4096


static int convert_image( unicap_data_buffer_t *data_buffer, AVFrame *frame, enum PixelFormat dest_pix_fmt )
{
   int src_pix_fmt = -1;
   int i;
   AVPicture src;
   AVPicture dest;
   
   for( i = 0; ( i < ucil_NumPixelFormatFourccs ) && ( src_pix_fmt == -1 ); i++ )
   {
      if( PixelFormatFourccs[i].fourcc == data_buffer->format.fourcc )
      {
	 src_pix_fmt = PixelFormatFourccs[i].pix_fmt;
      }
   }
   
   if( src_pix_fmt == -1 )
   {
      printf( "src_pix_fmt == -1 fourcc= %08x %c%c%c%c\n", data_buffer->format.fourcc, (data_buffer->format.fourcc>>24) & 0xff,
	      (data_buffer->format.fourcc>>16) & 0xff,(data_buffer->format.fourcc>>8) & 0xff,(data_buffer->format.fourcc) & 0xff);
      return 0;
   }
   
   src.data[0] = src.data[1] = src.data[2] = src.data[3] = data_buffer->data;
   // reassign src->data[x] here for packet formats
   src.linesize[0] = src.linesize[1] = src.linesize[2] = src.linesize[3] = data_buffer->format.size.width * data_buffer->format.bpp / 8;

   if( ( dest_pix_fmt < 0 ) || ( dest_pix_fmt > PIX_FMT_NB ) )
   {
      printf( "dest pix fmt out of range\n" );
      return 0;
   }
   avpicture_fill( (AVPicture*)&dest, frame->data[0], dest_pix_fmt, data_buffer->format.size.width, data_buffer->format.size.height );
   
   img_convert( &dest, dest_pix_fmt, &src, src_pix_fmt, data_buffer->format.size.width, data_buffer->format.size.height );   
   
   return 1;
}

static AVCodecContext *setup_codec( unicap_format_t *format, int codec_id, int pix_fmt )
{
   AVCodec *codec;
   AVCodecContext *c;

   codec = avcodec_find_encoder( codec_id );
   if( !codec )
   {
      fprintf( stderr, "Codec not found!\n" );
      return NULL;
   }
   
   c = avcodec_alloc_context( );
   c->bit_rate = 3000000;
   c->width = format->size.width;
   c->height = format->size.height;
   c->time_base = (AVRational){1,30};
   c->pix_fmt = pix_fmt;

   if( avcodec_open( c, codec ) < 0 )
   {
      fprintf( stderr, "Could not open codec\n" );
      return NULL;
   }
   
   
   return c;
}

static void encode_frame( ucil_mpeg_video_file_object_t *vobj )
{
   int out_size;
   
   out_size = avcodec_encode_video( vobj->ccontext, vobj->fbuf, vobj->fbuf_size, vobj->picture );
   
   fwrite( vobj->fbuf, 1, out_size, vobj->f );
}

static AVFrame *allocate_avframe( int width, int height, int pix_fmt )
{
   int bpp;
   int size;
   AVFrame *frame;
   unsigned char *framebuffer;
   
   
   if( ( pix_fmt < 0 ) || ( pix_fmt > PIX_FMT_NB ) )
   {
      return NULL;
   }
   
   bpp = PixelFormatBPP[pix_fmt];
   size = width * height * bpp / 8;
   
   frame = avcodec_alloc_frame( );
   framebuffer = malloc( avpicture_get_size( pix_fmt, width, height) );
   avpicture_fill( (AVPicture*)frame, framebuffer, pix_fmt, width, height );
   
   return frame;
}

static void finish_encode( ucil_mpeg_video_file_object_t *vobj )
{
   int out_size;
   
   for( out_size = 1; out_size; out_size = avcodec_encode_video( vobj->ccontext, vobj->fbuf, vobj->fbuf_size, NULL ) )
   {
      fwrite( vobj->fbuf, 1, out_size, vobj->f );
   }

   vobj->fbuf[0] = 0x0;
   vobj->fbuf[1] = 0x0;
   vobj->fbuf[2] = 0x1;
   vobj->fbuf[3] = 0xb7;
   
   fwrite( vobj->fbuf, 1, 4, vobj->f );
}

static unicap_status_t get_next_frame( ucil_mpeg_input_file_object_t vobj )
{
   int total_bytes_decoded = 0;
   
   while( 1 )
   {
      while( vobj->bytes_remain > 0 )
      {
	 int bytes_decoded;
	 int finished;
	 
	 bytes_decoded = avcodec_decode_video( vobj->pCodecCtx, vobj->pFrame, &finished, vobj->raw_data, vobj->bytes_remain );
	 if( bytes_decoded < 0 )
	 {
	    TRACE( "Failed to decode frame\n" );
	    return STATUS_FAILURE;
	 }
	 
	 TRACE( "finished: %d\n", finished );
	 

	 vobj->bytes_remain -= bytes_decoded;
	 vobj->raw_data += bytes_decoded;
	 
	 total_bytes_decoded += bytes_decoded;
	 
	 TRACE( "Total decoded: %d\n", total_bytes_decoded );

	 if( finished )
	 {
	    return STATUS_SUCCESS;
	 }
      }
      
      do
      {
	 if( vobj->packet.data != NULL )
	 {
	    av_free_packet( &vobj->packet );
	    vobj->packet.data = NULL;
	 }
	 
	 av_init_packet( &vobj->packet );
	 
	 if( av_read_packet( vobj->pFormatCtx, & vobj->packet ) < 0 )
	 {
	    int finished;
	    
	    avcodec_decode_video( vobj->pCodecCtx, vobj->pFrame, &finished, vobj->raw_data, vobj->bytes_remain );
	    
	    if( vobj->packet.data != NULL )
	    {
	       av_free_packet( &vobj->packet );
	       vobj->packet.data = NULL;
	    }
	    
	    return finished ? STATUS_SUCCESS : STATUS_FAILURE;
	 } 
      } while( vobj->packet.stream_index != vobj->video_stream );
      
      
      vobj->bytes_remain = vobj->packet.size;
      vobj->raw_data = vobj->packet.data;
      
   }
   
   return STATUS_FAILURE;
}

static void *mpegcpi_worker_thread( ucil_mpeg_input_file_object_t vobj )
{
   unicap_data_buffer_t new_frame_buffer;

   struct timeval ltime;

   vobj->packet.data = NULL;
   vobj->bytes_remain = 0;
   vobj->raw_data = NULL;
   
   unicap_copy_format( &new_frame_buffer.format, &vobj->format );
   new_frame_buffer.buffer_size = vobj->format.buffer_size;
   new_frame_buffer.type = UNICAP_BUFFER_TYPE_SYSTEM;

   gettimeofday( &ltime, NULL );
   
   while( !vobj->quit_capture_thread )
   {
      struct timespec abs_timeout;
      struct timeval  ctime;
      struct _unicap_queue *entry;
      

      gettimeofday( &ctime, NULL );
      abs_timeout.tv_sec = ctime.tv_sec + 1;
      abs_timeout.tv_nsec = ctime.tv_usec * 1000;      

      if( sem_timedwait( &vobj->sema, &abs_timeout ) )
      {
	 TRACE( "SEM_WAIT FAILED\n" );
	 continue;
      }

      if( !SUCCESS( get_next_frame( vobj ) ) )
      {
	 TRACE( "get_next failed!\n" );
      }
      

      img_convert( (AVPicture *)vobj->pOutFrame, vobj->out_pix_fmt, 
		   (AVPicture *)vobj->pFrame, vobj->in_pix_fmt, 
		   vobj->format.size.width, vobj->format.size.height );   
      
      if( vobj->event_callback )
      {
	 new_frame_buffer.data = vobj->pOutFrame->data[0];
	 vobj->event_callback( vobj->event_unicap_handle, UNICAP_EVENT_NEW_FRAME, &new_frame_buffer );
      }
      
/*       entry = _get_front_queue( vobj->in_queue ); */
/*       if( entry ) */
/*       { */
/* 	 unicap_data_buffer_t *data_buffer = ( unicap_data_buffer_t * ) entry->data; */

/* 	 unicap_copy_format( &data_buffer->format, &vobj->format ); */
/* 	 memcpy( data_buffer->data, vobj->pOutFrame->data[0], vobj->format.buffer_size ); */
	 
/* 	 _insert_back_queue( vobj->out_queue, entry );	  */
/*       } */
      
      gettimeofday( &ctime, NULL );
      if( ctime.tv_usec < ltime.tv_usec )
      {
	 ctime.tv_usec += 1000000;
	 ctime.tv_sec -= 1;
      }
      
      ctime.tv_usec -= ltime.tv_usec;
      ctime.tv_sec -= ltime.tv_sec;
      
      if( ( ctime.tv_sec == 0 ) &&
	  ( ctime.tv_usec < vobj->frame_intervall ) )
      {
	 usleep( vobj->frame_intervall - ctime.tv_usec );
      }
      
      gettimeofday( &ltime, NULL );

      sem_post( &vobj->sema );
      
   }
}

static unicap_status_t mpegcpi_reenumerate_formats( ucil_mpeg_input_file_object_t vobj, int *count )
{
   *count = 1;

   return STATUS_SUCCESS;
}

static unicap_status_t mpegcpi_enumerate_formats( ucil_mpeg_input_file_object_t vobj, unicap_format_t *format, int index )
{
   unicap_status_t status = STATUS_NO_MATCH;
   if( index == 0 )
   {
      unicap_copy_format( format, &vobj->format );
      status = STATUS_SUCCESS;
   }
   
   return status;
}

static unicap_status_t mpegcpi_set_format( ucil_mpeg_input_file_object_t vobj, unicap_format_t *format )
{
   unicap_status_t status = STATUS_SUCCESS;
   if( ( format->size.width != vobj->format.size.width ) || 
       ( format->size.height != vobj->format.size.height ) ||
       ( format->bpp != vobj->format.bpp ) )
   {
      char buffer[1024];
      int size = 1024;

      unicap_describe_format( format, buffer, &size );
      TRACE( "Could not set format: %s\n", buffer );
      size = 1024;
      unicap_describe_format( &vobj->format, buffer, &size );
      TRACE( "Stored: %s\n" );
      status = STATUS_FAILURE;
   }   
   
   return status;
}

static unicap_status_t mpegcpi_get_format( ucil_mpeg_input_file_object_t vobj, unicap_format_t *format )
{
   unicap_copy_format( format, &vobj->format );
   return STATUS_SUCCESS;
}

static unicap_status_t mpegcpi_reenumerate_properties( ucil_mpeg_input_file_object_t vobj, int *count )
{
   *count = 0;
   return STATUS_SUCCESS;
}

static unicap_status_t mpegcpi_enumerate_properties( ucil_mpeg_input_file_object_t vobj, unicap_property_t *property, int index )
{
   return STATUS_NO_MATCH;
}

static unicap_status_t mpegcpi_set_property( ucil_mpeg_input_file_object_t vobj, unicap_property_t *property )
{
   return STATUS_FAILURE;
}

static unicap_status_t mpegcpi_get_property( ucil_mpeg_input_file_object_t vobj, unicap_property_t *property )
{
   return STATUS_FAILURE;
}

static unicap_status_t mpegcpi_capture_start( ucil_mpeg_input_file_object_t vobj )
{
   unicap_status_t status = STATUS_SUCCESS;
   
   if( pthread_create( &vobj->worker_thread, NULL, (void*(*)(void*))mpegcpi_worker_thread, vobj ) )
   {
      TRACE( "Failed to create worker thread!\n" );
      return STATUS_FAILURE;
   }

   return status;
}

static unicap_status_t mpegcpi_capture_stop( ucil_mpeg_input_file_object_t vobj )
{
   unicap_status_t status = STATUS_SUCCESS;
   int res;

   vobj->quit_capture_thread = 1;
   res = pthread_join( vobj->worker_thread, NULL );

   return ( res == 0 ) ? STATUS_SUCCESS : STATUS_FAILURE;
}

static unicap_status_t mpegcpi_queue_buffer( ucil_mpeg_input_file_object_t vobj, unicap_data_buffer_t *buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   return status;
}

static unicap_status_t mpegcpi_dequeue_buffer( ucil_mpeg_input_file_object_t vobj, unicap_data_buffer_t **buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   return status;
}

static unicap_status_t mpegcpi_wait_buffer( ucil_mpeg_input_file_object_t vobj, unicap_data_buffer_t **buffer )
{
   unicap_status_t status = STATUS_SUCCESS;
   return status;
}

static unicap_status_t mpegcpi_poll_buffer( ucil_mpeg_input_file_object_t vobj, int *count )
{
   *count = 1;
   return STATUS_SUCCESS;
}

static unicap_status_t mpegcpi_set_event_notify( ucil_mpeg_input_file_object_t vobj, unicap_event_callback_t func, unicap_handle_t handle )
{
   vobj->event_callback = func;
   vobj->event_unicap_handle = handle;

   return STATUS_SUCCESS;
}


ucil_mpeg_video_file_object_t *ucil_mpeg_create_video_file( const char *path, unicap_format_t *format, ...)
{
   const int pix_fmt = PIX_FMT_YUV420P;
   ucil_mpeg_video_file_object_t *vobj = malloc( sizeof( struct _ucil_mpeg_video_file_object ) );   

   if( !initialized )
   {
      av_register_all();
      initialized = 1;
   }

   vobj->ccontext = setup_codec( format, CODEC_ID_MPEG1VIDEO, pix_fmt );
   if( !vobj->ccontext )
   {
      free( vobj );
      return 0;
   }
   
   vobj->picture = allocate_avframe( format->size.width, format->size.height, pix_fmt );
   if( !vobj->picture )
   {
      free( vobj );
      return 0;
   }
   
   vobj->f = fopen( path, "w" );
   if( !vobj->f )
   {
      av_free( vobj->picture );
      free( vobj );
      return 0;
   }
   
   vobj->fbuf_size = 1000000;
   vobj->fbuf = malloc( vobj->fbuf_size );
   vobj->pix_fmt = pix_fmt;
   
   return vobj;
}

unicap_status_t ucil_mpeg_encode_frame( ucil_mpeg_video_file_object_t *vobj, unicap_data_buffer_t *buffer )
{
   convert_image( buffer, vobj->picture, vobj->pix_fmt );
   encode_frame( vobj );
   
   return STATUS_SUCCESS;
}



unicap_status_t ucil_mpeg_close_video_file( ucil_mpeg_video_file_object_t *vobj )
{
   finish_encode( vobj );
   fclose( vobj->f );
   
   av_free( vobj->picture );
   free( vobj->fbuf );
   free( vobj );

   return STATUS_SUCCESS;
}

unicap_status_t ucil_mpeg_open_video_file( unicap_handle_t *unicap_handle, char *filename, unicap_format_t *format )
{
   struct _unicap_handle *handle;
   AVFormatContext *pFormatCtx;   
   AVCodecContext  *pCodecCtx;
   AVCodec         *pCodec;
   AVInputFormat   *iformat;
   AVProbeData      probedata;
   ucil_mpeg_input_file_object_t vobj;
   int i, videoStream;
   int pix_fmt;
   int ret;
   uint8_t inbuf[INBUF_SIZE + FF_INPUT_BUFFER_PADDING_SIZE];
   uint8_t *inbuf_ptr;

   if( !initialized )
   {
      av_register_all();
      initialized = 1;
   }


   /* set end of buffer to 0 (this ensures that no overreading happens for damaged mpeg streams) */
   memset(inbuf + INBUF_SIZE, 0, FF_INPUT_BUFFER_PADDING_SIZE);
   probedata.filename = filename;
   probedata.buf = inbuf;
   probedata.buf_size = INBUF_SIZE;
   iformat = av_probe_input_format( &probedata, 0 );
   

   if( ( ret = av_open_input_file( &pFormatCtx, filename, iformat, 0, NULL ) ) )
   {
      TRACE( "[%d] Could not open input file: '%s'\n", ret, filename );
      return STATUS_FAILURE;
   }

   if( av_find_stream_info( pFormatCtx ) < 0 )
   {
      TRACE( "Could not get stream info for file '%s'\n", filename );
      av_close_input_file( pFormatCtx );
      return STATUS_FAILURE;
   }

   for( i = 0; i < pFormatCtx->nb_streams; i++ )
   {
      if( pFormatCtx->streams[ i ]->codec->codec_type == CODEC_TYPE_VIDEO  )
      {
	 videoStream = i;
	 break;
      }
   }
   if( videoStream == -1 )
   {
      TRACE( "Could not find a video stream in file '%s'\n", filename );
      av_close_input_file( pFormatCtx );
      return STATUS_FAILURE;
   }

   handle = malloc( sizeof( struct _unicap_handle ) );
   if( !handle )
   {
      av_close_input_file( pFormatCtx );
      return STATUS_FAILURE;
   }
   // initialize data structure
   memset( handle, 0, sizeof( struct _unicap_handle ) );
   handle->cb_info = malloc( sizeof( struct _unicap_callback_info ) * UNICAP_EVENT_LAST );
   for( i = 0; i < UNICAP_EVENT_LAST; i++ )
   {
      handle->cb_info[ i ].func = NULL;
      handle->cb_info[ i ].user_ptr = NULL;
   }

   handle->sem_key = -1;
   handle->sem_id = -1;
   
   handle->ref_count = malloc( sizeof( int ) );
   *(handle->ref_count) = 1;

   vobj = malloc( sizeof( struct _ucil_mpeg_input_file_object ) );
   if( !vobj )
   {
      return STATUS_FAILURE;
   }   

   memset( vobj, 0x0, sizeof( struct _ucil_mpeg_input_file_object ) );

   pCodecCtx = pFormatCtx->streams[ videoStream ]->codec;   
   pCodec = avcodec_find_decoder( pCodecCtx->codec_id );
   if( pCodec == NULL )
   {
      TRACE( "No decoder found for file '%s'\n", filename );
      return STATUS_FAILURE;
   }
   
   if( pCodec->capabilities & CODEC_CAP_TRUNCATED )
   {
      pCodecCtx->flags |= CODEC_FLAG_TRUNCATED;
   }

   if(avcodec_open(pCodecCtx, pCodec)<0)
   {
      TRACE( "Could not open codec for file '%s'\n", filename );
      return STATUS_FAILURE;
   }
      
   if( ( pCodecCtx->width != format->size.width ) || 
       ( pCodecCtx->height != format->size.height ) )
   {
      TRACE( "Wrong size for format: width = %d <> %d; height = %d <> %d\n", 
	     pCodecCtx->width, format->size.width, pCodecCtx->height, format->size.height ) ;
      
      return STATUS_FAILURE;
   }
   
   vobj->pFrame = avcodec_alloc_frame();
   
   pix_fmt = -1;
   for( i = 0; ( i < ucil_NumPixelFormatFourccs ) && ( pix_fmt == -1 ); i++ )
   {
      if( PixelFormatFourccs[i].fourcc == format->fourcc )
      {
	 pix_fmt = PixelFormatFourccs[i].pix_fmt;
      }
   }
   
   vobj->pOutFrame = allocate_avframe( format->size.width, format->size.height, pix_fmt );   
   vobj->out_pix_fmt = pix_fmt;
   vobj->in_pix_fmt = pCodecCtx->pix_fmt;

   vobj->video_stream = videoStream;
   vobj->pCodec = pCodec;
   vobj->pCodecCtx = pCodecCtx;
   vobj->pFormatCtx = pFormatCtx;
   vobj->quit_capture_thread = 0;
   vobj->frame_intervall = 30000;

   unicap_copy_format( &vobj->format, format );   
   strcpy( vobj->format.identifier, "UCIL default format" );
   vobj->format.buffer_size = vobj->format.size.width * vobj->format.size.height * vobj->format.bpp / 8;
   vobj->format.buffer_type = UNICAP_BUFFER_TYPE_SYSTEM;

   handle->cpi.cpi_register = NULL;
   handle->cpi.cpi_version = 1<<16;
   handle->cpi.cpi_capabilities = 0x3ffff;
   handle->cpi.cpi_enumerate_devices = NULL;
   handle->cpi.cpi_open = NULL;
   handle->cpi.cpi_close = NULL;
   handle->cpi.cpi_reenumerate_formats = (cpi_reenumerate_formats_t ) mpegcpi_reenumerate_formats;
   handle->cpi.cpi_enumerate_formats = ( cpi_enumerate_formats_t ) mpegcpi_enumerate_formats;
   handle->cpi.cpi_set_format = ( cpi_set_format_t ) mpegcpi_set_format;
   handle->cpi.cpi_get_format = ( cpi_get_format_t ) mpegcpi_get_format;
   handle->cpi.cpi_reenumerate_properties = ( cpi_reenumerate_properties_t ) mpegcpi_reenumerate_properties;
   handle->cpi.cpi_enumerate_properties = ( cpi_enumerate_properties_t ) mpegcpi_enumerate_properties;
   handle->cpi.cpi_get_property = ( cpi_get_property_t ) mpegcpi_get_property;
   handle->cpi.cpi_set_property = ( cpi_set_property_t ) mpegcpi_set_property;
   handle->cpi.cpi_capture_start = ( cpi_capture_start_t ) mpegcpi_capture_start;
   handle->cpi.cpi_capture_stop = ( cpi_capture_stop_t ) mpegcpi_capture_stop;
   handle->cpi.cpi_queue_buffer = ( cpi_queue_buffer_t ) mpegcpi_queue_buffer;
   handle->cpi.cpi_dequeue_buffer = ( cpi_dequeue_buffer_t ) mpegcpi_dequeue_buffer;
   handle->cpi.cpi_wait_buffer = ( cpi_wait_buffer_t ) mpegcpi_wait_buffer;
   handle->cpi.cpi_poll_buffer = ( cpi_poll_buffer_t ) mpegcpi_poll_buffer;
   handle->cpi.cpi_set_event_notify = ( cpi_set_event_notify_t ) mpegcpi_set_event_notify;
   handle->cpi_data = vobj;

   unicap_cpi_register_event_notification( mpegcpi_set_event_notify, vobj, handle );
   

   sem_init( &vobj->sema, 0, 1 );
      

   *unicap_handle = handle;

   return STATUS_SUCCESS;
}
